/********************************************************************************
 * Copyright (C) 2020 Genuitec, LLC. All rights reserved.
 *
 * Usage of this software is governed by the CodeTogether End User License Agreement.
 *
 * SPDX-License-Identifier: SEE LICENSE IN LICENSE
 ********************************************************************************/

import { inject, injectable } from 'inversify';
import { PresenceMemento } from '../codetogether-startup/presence-memento';
import { FrontendApplicationContribution } from '../frontend-application';
import { KeybindingsStyle } from '../ct-keybindings/ct-ide-keybindings';
import { ThemeService, BuiltinThemeProvider, Theme } from '../theming';
import { IDE } from '../codetogether-startup/session';
import { ColorApplicationContribution } from '../color-application-contribution';

export const THEME_IDE_KEY = 'theme_ide_style';

export const STARTUP_BACKGROUND_KEY = 'theme.background';

export enum ThemeIDE {
    EclipseLight = 'Eclipse Light',
    EclipseDark = 'Eclipse Dark',
    IntelliJLight = 'IntelliJ Light',
    IntelliJDark = 'IntelliJ Dark',
    VSCodeLight = 'VS Code Light',
    VSCodeDark = 'VS Code Dark',
    NoTheme = 'no-theme'
}

export enum StartUpBackgroundColors {
    Light = '#FFFFFF',
    Dark = '#1E1E1E'
}

export function themeIDEToTheiaTheme(themeIDE: ThemeIDE | undefined): string {
    switch (themeIDE) {
        case ThemeIDE.EclipseLight:
            return BuiltinThemeProvider.eclipseLightTheme.id;
        case ThemeIDE.EclipseDark:
            return BuiltinThemeProvider.eclipseDarkTheme.id;
        case ThemeIDE.IntelliJLight:
            return BuiltinThemeProvider.intelliJLightTheme.id;
        case ThemeIDE.IntelliJDark:
            return BuiltinThemeProvider.intelliJDarkTheme.id;
        case ThemeIDE.VSCodeLight:
            return BuiltinThemeProvider.lightTheme.id;
        default:
            return BuiltinThemeProvider.darkTheme.id;
    }
}

export function theiaThemeToThemeIDE(theiaTheme: Theme): ThemeIDE {
    switch (theiaTheme.id) {
        case BuiltinThemeProvider.eclipseLightTheme.id:
            return ThemeIDE.EclipseLight;
        case BuiltinThemeProvider.eclipseDarkTheme.id:
            return ThemeIDE.EclipseDark;
        case BuiltinThemeProvider.intelliJLightTheme.id:
            return ThemeIDE.IntelliJLight;
        case BuiltinThemeProvider.intelliJDarkTheme.id:
            return ThemeIDE.IntelliJDark;
        case BuiltinThemeProvider.lightTheme.id:
            return ThemeIDE.VSCodeLight;
        default:
            return ThemeIDE.VSCodeDark;
    }
}

export function ideToDefaultThemeIDE(ide: IDE): ThemeIDE {
    switch (ide) {
        case IDE.Eclipse:
            return ThemeIDE.EclipseLight;
        case IDE.IntelliJ:
            return ThemeIDE.IntelliJLight;
        default:
            return ThemeIDE.VSCodeDark;
    }
}

export function themeIDEtoKeybinding(themeIDE: ThemeIDE | undefined): KeybindingsStyle {
    switch (themeIDE) {
        case ThemeIDE.EclipseDark:
            return KeybindingsStyle.Eclipse;
        case ThemeIDE.EclipseLight:
            return KeybindingsStyle.Eclipse;
        case ThemeIDE.IntelliJDark:
            return KeybindingsStyle.IntelliJ;
        case ThemeIDE.IntelliJLight:
            return KeybindingsStyle.IntelliJ;
        case ThemeIDE.VSCodeDark:
            return KeybindingsStyle.VSCode;
        case ThemeIDE.VSCodeLight:
            return KeybindingsStyle.VSCode;
        default:
            return KeybindingsStyle.Eclipse;
    }
}

export function isLightTheme(themeIDE: ThemeIDE): boolean {
    const themeId = themeIDEToTheiaTheme(themeIDE);
    return themeId.includes('light');
}

@injectable()
export class ThemeIDEService implements FrontendApplicationContribution {
    constructor(
        @inject(PresenceMemento) protected readonly presenceMemento: PresenceMemento,
        @inject(ThemeService) protected readonly themeService: ThemeService
    ) { }

    setThemeIDE(themeIDE: ThemeIDE, updateStartupBackground = false): void {
        this.themeService.setCurrentTheme(
            themeIDEToTheiaTheme(themeIDE),
        );
        this.storeThemeIDE(themeIDE);
        if (updateStartupBackground) {
            this.updateStartupBackground(themeIDE);
        }
    }

    getThemeIDE(): Promise<ThemeIDE | undefined> {
        return Promise.resolve(this.presenceMemento.get(THEME_IDE_KEY));
    }

    storeThemeIDE(style: ThemeIDE | undefined): Promise<void> {
        return this.presenceMemento.update(THEME_IDE_KEY, style);
    }

    updateStartupBackground(theme: ThemeIDE): void {
        const backgroundColor = isLightTheme(theme) ? StartUpBackgroundColors.Light : StartUpBackgroundColors.Dark;
        window.localStorage.setItem(STARTUP_BACKGROUND_KEY, backgroundColor);
        ColorApplicationContribution.initBackground();
    }

    async onStart(): Promise<void> {
        this.themeService.onThemeChange(event => {
            this.storeThemeIDE(theiaThemeToThemeIDE(event.newTheme));
        });

        const themeIDE = await this.getThemeIDE();
        if (themeIDE) {
            this.setThemeIDE(themeIDE);
        }
    }
}
